/*
 * Copyright (c) 2022-present Charles7c Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.zayac.admin.schedule;

import com.zayac.admin.req.team.TeamInfoReq;
import com.zayac.admin.resp.team.Team;
import com.zayac.admin.service.*;
import com.zayac.admin.system.model.resp.AccountResp;
import com.zayac.admin.system.model.resp.DeptUsersResp;
import com.zayac.admin.system.model.resp.UserWithRolesAndAccountsResp;
import com.zayac.admin.system.service.*;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.context.annotation.Profile;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.temporal.TemporalAdjusters;
import java.util.AbstractMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.Executor;
import java.util.stream.Collectors;

/**
 * 检测注册跟新增的定时任务
 */

@Slf4j
@Component
@RequiredArgsConstructor
@Profile("prod")
public class CheckRegAndDep {

    private final TeamService teamService;
    private final DeptService deptService;
    private final RegistrationService registrationService;
    private final DepositService depositService;
    private final Executor asyncTaskExecutor;

    private static final String MINISTER_ROLE_CODE = "minister";
    private static final long FIXED_DELAY = 60000L;

    @Scheduled(fixedDelay = FIXED_DELAY)
    public void checkRegistrationAndNewDeposit() {
        LocalDate nowDate = LocalDate.now();
        LocalDateTime nowDateTime = LocalDateTime.now();
        //查询用户角色为部长的部门所有用户
        List<DeptUsersResp> deptWithUsersAndAccounts = deptService.getDeptWithUsersAndAccounts(MINISTER_ROLE_CODE);
        deptWithUsersAndAccounts.forEach(dept -> {
            //根据用户角色对部门用户进行分组
            Map<String, List<UserWithRolesAndAccountsResp>> usersByRole = dept.getUsers()
                .stream()
                .flatMap(user -> user.getRoles()
                    .stream()
                    .map(role -> new AbstractMap.SimpleEntry<>(role.getCode(), user)))
                .collect(Collectors.groupingByConcurrent(Map.Entry::getKey, Collectors
                    .mapping(Map.Entry::getValue, Collectors.toList())));
            // 获取所有账号的username与用户的映射
            Map<String, UserWithRolesAndAccountsResp> accountUsernameToUserMap = dept.getUsers()
                .stream()
                .flatMap(user -> user.getAccounts()
                    .stream()
                    .map(account -> new AbstractMap.SimpleEntry<>(account.getUsername(), user)))
                .collect(Collectors.toConcurrentMap(Map.Entry::getKey, Map.Entry::getValue));

            var ministerUser = usersByRole.get(MINISTER_ROLE_CODE).get(0);
            processUser(ministerUser, accountUsernameToUserMap, nowDate, nowDateTime).join();
        });

    }

    private CompletableFuture<Void> processUser(UserWithRolesAndAccountsResp minister,
                                                Map<String, UserWithRolesAndAccountsResp> accountUsernameToUserMap,
                                                LocalDate nowDate,
                                                LocalDateTime nowDateTime) {
        //根据总线用户的账号查询数据
        List<AccountResp> accounts = minister.getAccounts();
        List<CompletableFuture<Void>> futures = accounts.stream()
            .map(account -> processTeamAccount(minister, accountUsernameToUserMap, account, nowDate, nowDateTime))
            .toList();

        return CompletableFuture.allOf(futures.toArray(new CompletableFuture[0]));
    }

    private CompletableFuture<Void> processTeamAccount(UserWithRolesAndAccountsResp minister,
                                                       Map<String, UserWithRolesAndAccountsResp> accountUsernameToUserMap,
                                                       AccountResp account,
                                                       LocalDate nowDate,
                                                       LocalDateTime nowDateTime) {
        TeamInfoReq teamInfoReq = TeamInfoReq.builder()
            .startDate(nowDateTime.with(TemporalAdjusters.firstDayOfMonth()).with(LocalTime.MIN))
            .endDate(nowDateTime.with(TemporalAdjusters.lastDayOfMonth()).with(LocalTime.MAX))
            .build();
        CompletableFuture<Team> teamFuture = teamService.getLatestTeamInfoAsync(account, teamInfoReq);
        Team prevTeamInfo = teamService.getPreviousTeamInfo(account);

        return teamFuture.thenComposeAsync(currentTeamInfo -> {
            log.info("Previous Team Info: {}", prevTeamInfo);
            log.info("Current Team Info: {}", currentTeamInfo);

            CompletableFuture<Void> registrationProcess = registrationService
                .processRegistration(minister, account, accountUsernameToUserMap, currentTeamInfo, prevTeamInfo, nowDate, asyncTaskExecutor);

            CompletableFuture<Void> depositProcess = depositService
                .processDeposits(minister, accountUsernameToUserMap, account, currentTeamInfo, prevTeamInfo, nowDate, asyncTaskExecutor);

            return CompletableFuture.allOf(registrationProcess, depositProcess).thenRunAsync(() -> {
                teamService.updateTeamInfo(account, currentTeamInfo);
                log.info("Team info updated");
            }, asyncTaskExecutor);
        }, asyncTaskExecutor);
    }
}
