/*
 * Copyright (c) 2022-present Charles7c Authors. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.zayac.admin.service;

import cn.hutool.core.lang.TypeReference;
import cn.hutool.json.JSONUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.util.concurrent.RateLimiter;
import com.zayac.admin.config.RateLimiterConfig;
import com.zayac.admin.resp.ApiResponse;
import com.zayac.admin.system.model.resp.AccountResp;
import io.netty.handler.timeout.TimeoutException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.http.HttpStatusCode;
import org.springframework.stereotype.Service;
import org.springframework.web.reactive.function.BodyInserters;
import org.springframework.web.reactive.function.client.WebClient;
import org.springframework.web.reactive.function.client.WebClientRequestException;
import org.springframework.web.reactive.function.client.WebClientResponseException;
import reactor.core.publisher.Mono;
import reactor.core.scheduler.Schedulers;
import reactor.util.retry.Retry;
import top.continew.starter.core.exception.BusinessException;

import java.time.Duration;
import java.util.Map;
import java.util.concurrent.CompletableFuture;

@Service
@Slf4j
public class CompletableFutureWebClientService {
    private final WebClient webClient;
    private final RateLimiterConfig rateLimiterConfig;
    private final ObjectMapper objectMapper;

    public CompletableFutureWebClientService(WebClient.Builder webClientBuilder,
                                             RateLimiterConfig rateLimiterConfig,
                                             ObjectMapper objectMapper) {
        this.webClient = webClientBuilder.build();
        this.rateLimiterConfig = rateLimiterConfig;
        this.objectMapper = objectMapper;
    }

    public <T> CompletableFuture<T> fetchDataForAccount(AccountResp account,
                                                        String apiPath,
                                                        Object params,
                                                        ParameterizedTypeReference<ApiResponse<T>> typeRef) {
        RateLimiter rateLimiter = rateLimiterConfig.getRateLimiter(account.getPlatformUrl() + apiPath);
        return fetchData(rateLimiter, account.getPlatformUrl() + apiPath, account
            .getHeaders(), params, typeRef, account).toFuture();
    }

    public <T> Mono<T> fetchData(RateLimiter rateLimiter,
                                 String url,
                                 String headers,
                                 Object params,
                                 ParameterizedTypeReference<ApiResponse<T>> typeRef,
                                 AccountResp account) {
        return Mono.fromCallable(() -> {
            rateLimiter.acquire();
            return true;
        })
            .subscribeOn(Schedulers.boundedElastic())
            .flatMap(ignored -> webClient.post().uri(url).headers(httpHeaders -> {
                addHeaders(httpHeaders, headers);
            })
                .body(params != null ? BodyInserters.fromValue(params) : BodyInserters.empty())
                .retrieve()
                .onStatus(HttpStatusCode::isError, response -> Mono.error(new BusinessException("API call failed")))
                .bodyToMono(String.class)
                .doOnNext(resStr -> {
                    log.debug("Request URL: {}", url);
                    log.debug("Request headers: {}", headers);
                    log.debug("Request params: {}", params);
                    log.debug("Response: {}", resStr);
                })
                .flatMap(body -> {
                    try {
                        ApiResponse<T> apiResponse = objectMapper.readValue(body, objectMapper.getTypeFactory()
                            .constructType(typeRef.getType()));
                        return Mono.justOrEmpty(apiResponse);
                    } catch (Exception e) {
                        log.warn("JSON parsing exception: " + e.getMessage());
                        return Mono.just(new ApiResponse<T>(null, "Decoding error", 6008));
                    }
                })
                .flatMap(response -> respHandler(response, account))
                .retryWhen(Retry.backoff(3, Duration.ofSeconds(5)).filter(this::isRetryableException)));
    }

    private boolean isRetryableException(Throwable throwable) {
        return throwable instanceof TimeoutException || throwable instanceof WebClientResponseException.ServiceUnavailable || throwable instanceof WebClientRequestException;
    }

    private void addHeaders(org.springframework.http.HttpHeaders httpHeaders, String headers) {
        try {
            Map<String, String> headerMap = JSONUtil.toBean(headers, new TypeReference<Map<String, String>>() {
            }, true);
            headerMap.forEach(httpHeaders::add);
        } catch (Exception e) {
            log.warn("Header conversion exception: " + e.getMessage());
            throw new BusinessException("Header conversion failed", e);
        }
    }

    private <T> Mono<T> respHandler(ApiResponse<T> response, AccountResp account) {
        if (response.getStatusCode().equals(6000)) {
            return Mono.just(response.getData());
        } else if (response.getStatusCode().equals(6001)) {
            log.warn("[{}]登录失效,请检查token有效性", account.getUsername());
            throw new BusinessException("登录失效,请检查token有效性");
        } else {
            return Mono.error(new BusinessException("Error status code: " + response.getStatusCode()));
        }
    }
}
